from sklearn.neural_network import BernoulliRBM

import os
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
from chembee_config.benchmark.BenchmarkAlgorithm import BenchmarkAlgorithm


class RBMClassifier(BenchmarkAlgorithm):

    name = "restritce-bm"

    algorithms = (
        BernoulliRBM(n_components=256, learning_rate=0.0001),
        BernoulliRBM(n_components=612, learning_rate=0.0001),
        BernoulliRBM(n_components=256, learning_rate=0.000001),
        BernoulliRBM(n_components=612, learning_rate=0.000001),
    )
    titles = (
        "256 units, 0.01 lr",
        "612 units, 0.01 lr",
        "256 units, 0.001 lr",
        "612 units, 0.001 lr",
    )
    _response_method = "transform"



import sys, os
import numpy as np

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
from chembee_config.benchmark.svc import SVClassifier
from chembee_config.benchmark.svc_poly import SVCPolyClassifier
from chembee_config.benchmark.spectral_clustering import SpectralClusteringClassifier
from chembee_config.benchmark.random_forest import RandomForestClassifier
from chembee_config.benchmark.naive_bayes import NaiveBayesClassifier
from chembee_config.benchmark.logistic_regression import LogisticRegressionClassifier
from chembee_config.benchmark.linear_regression import LinearRegressionClassifier
from chembee_config.benchmark.kmeans import KMeansClassifier
from chembee_config.benchmark.knn import KNNClassifier
from chembee_config.benchmark.mlp_classifier import NeuralNetworkClassifier
from chembee_config.benchmark.restricted_bm import RBMClassifier

from chembee_plotting.graphics import plotting_map_comparison
from file_utils import save_json_to_file, make_full_filename

import logging


logging.basicConfig(
    format="%(levelname)s:%(asctime)s %(message)s",
    datefmt="%m/%d/%Y %I:%M:%S %p",
    level=logging.INFO,
    filename="chembee.log",
)

algorithms = [
    SVClassifier,
    SVCPolyClassifier,
    SpectralClusteringClassifier,
    RandomForestClassifier,
    NaiveBayesClassifier,
    LogisticRegressionClassifier,
    LinearRegressionClassifier,
    KMeansClassifier,
    KNNClassifier,
    NeuralNetworkClassifier,
    RBMClassifier,
]


def benchmark_standard(
    X,
    y,
    feature_names=["Feature 1", "Feature 2"],
    file_name="benchmark",
    prefix="plots/benchmarks",
    algorithms=algorithms,
    to_fit=True,
):

    metrics = {}
    for alg in algorithms:
        if alg is None:
            response_method = alg._response_method
        else:
            response_method = None
        try:
            metric = benchmark_algorithm_standard(
                alg,
                X,
                y,
                plot_function=plotting_map_comparison[len(alg.algorithms)],
                file_name="benchmark_" + alg.name,
                prefix=prefix,
                feature_names=feature_names,
                response_method=None,
                to_fit=to_fit,
            )
        except Exception as e:
            assert 1 == 2, str(e)
            logging.info("Could not fit classifier " + str(alg.name))
            logging.info(str(e))
            metric = None
        finally:
            continue
    file_name = make_full_filename(file_name=file_name, prefix=prefix)
    save_json_to_file(metrics, file_name=file_name)
    return metrics


def benchmark_cv_algorithms(
    algorithms: list,
    names: list,
    X: np.ndarray,
    y: np.ndarray,
    plot_function=plotting_map_comparison[1],
    file_name: str = "after_cv_benchmark",
    prefix: str = "plots/benchmarks/",
    feature_names: list = ["Feature 1", "Feature 2"],
    response_method: str = "predict",
    to_fit=True,
):
    """
    The benchmark_cv_algorithms function takes a list of algorithms and fits them to the data.
    It then plots the results using plot_cv_algorithms function.

    :param algorithms:list: Used to specify which algorithms to use in the benchmark.
    :param names:list: Used to specify the names of the algorithms.
    :param X:np.ndarray: Used to pass the data to be used for training and testing.
    :param y:np.ndarray: Used to pass the target variable.
    :param file_name:str="benchmark": Used to name the output file.
    :param prefix:str="benchmarks/": Used to specify the path to the directory where all benchmark plots are saved.
    :param feature_names:list=["Feature1": Used to label the x-axis of the plots.
    :param "Feature2"]: Used to Specify the name of the feature that is used for plotting.
    :param response_method:str="predict": Used to determine which method is used to generate the response.
    :param to_fit=True: Used to determine whether the model should be fitted before predicting or transforming.
    :param : Used to determine the method used to obtain the response from the model.
    :return: A list of metrics for each model.

    :doc-author: Julian M. Kleber
    """

    metrics_class = []
    models = []

    for i in range(len(algorithms)):

        clf = algorithms[i]
        name = names[i]
        file_name = name + file_name
        if to_fit:
            try:
                clf.fit(X, y)
            except:
                logging.info("Could not fit clf " + str(name))
        models.append(clf)
        if response_method == "predict":
            try:
                y_pred = clf.predict(X)
            except:
                logging.info("Could not predict results for" + str(name))
        elif response_method == "fit_predict":
            try:
                y_pred = clf.fit_predict(X, y)
            except:
                logging.info("Could not fit_predict results for" + str(name))
        elif response_method == "transform":
            try:
                y_pred = clf.transform(X)
            except:
                logging.info("Could not transform results for" + str(name))
        logging.info("Fitted model: " + str(name))

        try:
            plot_function(
                clf=clf,
                X=X,
                y=y,
                file_name=file_name,
                prefix=prefix,
                feature_names=feature_names,
                response_method=response_method,
            )
        except:
            logging.info("Could not plot name " + str(name))

    metrics = {k: v for k, v in zip(names, metrics_class)}
    return metrics


def benchmark_algorithm_standard(
    algorithm: list,
    X: np.ndarray,
    y: np.ndarray,
    plot_function: object(),
    file_name: str = "benchmark",
    prefix: str = "benchmarks/",
    feature_names: list = ["Feature 1", "Feature 2"],
    response_method: str = "predict",
    to_fit=True,
) -> dict:
    """
    The benchmark_algorithm_standard function takes a list of algorithms,
    a list of names for the algorithms, and a dataset (X and y),
    and returns an array of metrics. The function is meant to be used for standard algorithms and not cross validation
    objects. For cross_validation objects, choose the benchmark_cv object. The function also plots the results.

    :param algorithm:list: Used to store the algorithms that will be used in the benchmark.
    :param names:list: Used to pass the names of the algorithms that will be used in the plot.
    :param X:np.ndarray: Used to pass the data to the benchmark_algorithm_standard function.
    :param y:np.ndarray: Used to pass the target variable to the plot_roc_curve function.
    :param plot_function:object(): Used to pass a function to the benchmark_algorithm_standard function.
    :param file_name:str="benchmark": Used to define the name of the file that will be saved.
    :param prefix:str="benchmarks/": Used to specify the path where the plots will be saved.
    :param feature_names:list=["Feature1": Used to label the x-axis of the plot.
    :param "Feature2"]: Used to Specify the name of the feature that is used in the plot.
    :param response_method:str="predict": Used to determine the method used to generate a response from the model.
    :param to_fit=True: Used to tell the algorithm to fit the model before predicting.
    :param : Used to define the type of plot to be created.
    :return: A dictionary of metrics.

    :doc-author: Julian M. Kleber"""

    name = algorithm.name
    algorithms = algorithm.algorithms
    file_name = "benchmark_" + name
    metrics_class = []
    models = []
    for i in range(len(algorithm.titles)):

        clf = algorithms[i]
        if to_fit:
            clf = clf.fit(X, y)
        models.append(clf)
        if response_method == "predict":
            y_pred = clf.predict(X)
        elif response_method == "fit_predict":
            y_pred = clf.fit_predict(X, y)
        if response_method == "transform":
            y_pred = clf.transform(X)
        logging.info("Fitted model: " + str(algorithm.titles[i]))

    plot_function(
        models=models,
        titles=algorithm.titles,
        X=X,
        y=y,
        file_name=file_name,
        prefix=prefix,
        feature_names=feature_names,
        response_method=response_method,
    )

    metrics = {k: v for k, v in zip(algorithm.titles, metrics_class)}
    return metrics


# should refactor the above functions
def benchmark_algorithm(
    algorithms: list,
    X: np.ndarray,
    y: np.ndarray,
    plot_function: object(),
    file_name: str = "benchmark",
    prefix: str = "benchmarks/",
    feature_names: list = ["Feature 1", "Feature 2"],
    response_method: str = "predict",
    to_fit=True,
) -> dict:
    """
    This function is not used in production at the moment and serves as a draft on how to abstract the functions above,
    to refactor the code and make it more maintainable.
    The benchmark_algorithm function is used to benchmark the performance of a given algorithm.
    It takes as input an algorithm, and returns a dictionary containing the metrics for that algorithm.
    The function is designed to be called within a loop, so that it can be used to compare multiple algorithms at once.

    :param algorithm: Used to specify the algorithm to use.
    :param X:np.ndarray: Used to pass the data to be used for training and testing.
    :param y:np.ndarray: Used to pass the target variable to the plot_roc_curve function.
    :param plot_function:object(): Used to pass a function that plots the results.
    :param file_name:str="benchmark": Used to name the plot file.
    :param prefix:str="benchmarks/": Used to specify the path to store the generated plots.
    :param feature_names:list=["Feature1": Used to name the columns of the dataframe.
    :param "Feature2"]: Used to Specify the names of the features in your dataset.
    :param response_method:str="predict": Used to determine whether the models should be used to predict the response or transform it.
    :param to_fit=True: Used to fit the model to the data.
    :param : Used to specify the name of the file to which we want to save our plots.
    :return: A dictionary with the following keys:.

    :doc-author: Julian M. Kleber
    """

    clf = algorithms[i]
    if to_fit:
        clf = clf.fit(X, y)
    if response_method == "predict":
        y_pred = clf.predict(X)
    elif response_method == "fit_predict":
        y_pred = clf.fit_predict(X, y)
    elif response_method == "transform":
        y_pred = clf.transform(X)
    else:
        logging.info("No valid response method")
    logging.info("Fitted model: " + str(algorithm.titles[i]))

    metrics = {k: v for k, v in zip(algorithm.titles, metrics_class)}
    return metrics



import os
from sklearn.metrics import accuracy_score, confusion_matrix, roc_curve, auc
from sklearn.model_selection import StratifiedKFold
import numpy as np

from chembee_actions.classifier_fit import clf_fit
from chembee_actions.clf_list import clf_list  # load classifiers

import logging

logging.basicConfig(
    format="%(levelname)s:%(asctime)s %(message)s",
    datefmt="%m/%d/%Y %I:%M:%S %p",
    level=logging.DEBUG,
    filename=os.getenv("chembee.log"),
)


def screen_classifier_for_metrics_stratified(
    X_data,
    y_true,
    n,
    file_name="evaluation",
    prefix="plots/evaluation",
    clf_list: list = clf_list,
) -> dict:
    """
    The screen_classifier_for_metrics function takes a list of classifiers and fits them to the training data using stratified n-fold cross validation.
    The goal is to have statistically more meaningful evaluations of a classifier and at the same time some sense of the distribution and purity of the
    dataset.
    The fubction predicts the test data using each classifier, calculates metrics for each prediction, and stores all
    of this information in a dictionary. The function returns this dictionary. Intended use for web application

    :param clf:list=clf_list: Used to pass the classifiers to be evaluated.
    :param X_train: Used to pass the training data set.
    :param y_train: Used to train the classifier.
    :param name: Used to create a folder where the plots will be stored.
    :param X_test: Used to test the classifier on a set of data.
    :param y_test: Used to calculate the metrics.
    :param file_name="evaluation": Used to specify the name of the file where all metrics will be saved.
    :param prefix="plots/evaluation": Used to set the path where the plots will be saved.
    :return: A dictionary of metrics for each classifier.

    :doc-author: Julian M. Kleber
    """

    cv = StratifiedKFold(n_splits=n)
    accuracies_test = []
    accuracies_train = []
    metrics_collection = init_metrics_collection()
    for clf in clf_list:
        metrics_collection["scalar"][clf.name] = {}
        metrics_collection["array"][clf.name] = {}
        metrics_collection["matrix"][clf.name] = {}
        for i, (train, test) in enumerate(cv.split(X_data, y_true.astype(np.int32))):
            name = clf.name
            clf, y_pred = clf_fit(
                clf, X_data[train], y_true.astype(np.int32)[train], name
            )
            y_pred = clf.predict(X_data[test])
            metric_scalar, metric_array, metric_matrix = calculate_metrics_classifier(
                clf, X_data[test], y_true[test], y_pred
            )
            metrics_collection["scalar"][clf.name][i] = metric_scalar
            metrics_collection["array"][clf.name][i] = metric_array
            metrics_collection["matrix"][clf.name][i] = metric_matrix

        metrics_collection["scalar"][clf.name] = parse_multi_output(
            metrics_collection["scalar"][clf.name]
        )
        metrics_collection["array"][clf.name] = parse_multi_output(
            metrics_collection["array"][clf.name]
        )
    return metrics_collection


def screen_classifier_for_metrics(
    X_train,
    y_train,
    X_test,
    y_test,
    file_name="evaluation",
    prefix="plots/evaluation",
    clf_list: list = clf_list,
    to_fit=True,
) -> dict:
    """
    The screen_classifier_for_metrics function takes a list of classifiers and fits them to the training data.
    It then predicts the test data using each classifier, calculates metrics for each prediction, and stores all
    of this information in a dictionary. The function returns this dictionary. Intended use for web application

    :param clf:list=clf_list: Used to pass the classifiers to be evaluated.
    :param X_train: Used to pass the training data set.
    :param y_train: Used to train the classifier.
    :param name: Used to create a folder where the plots will be stored.
    :param X_test: Used to test the classifier on a set of data.
    :param y_test: Used to calculate the metrics.
    :param file_name="evaluation": Used to specify the name of the file where all metrics will be saved.
    :param prefix="plots/evaluation": Used to set the path where the plots will be saved.
    :return: A dictionary of metrics for each classifier.

    :doc-author: Julian M. Kleber
    """
    metrics_collection = init_metrics_collection()

    for i in range(len(clf_list)):
        clf = clf_list[i]
        name = clf.name
        if to_fit:
            clf, y_pred = clf_fit(clf, X_train, y_train, name)
        y_pred = clf.predict(X_test).astype(np.int32)
        metric_scalar, metric_array, metric_matrix = calculate_metrics_classifier(
            clf, X_test, y_test, y_pred
        )
        metrics_collection["scalar"][clf.name] = metric_scalar
        metrics_collection["array"][clf.name] = metric_array
        metrics_collection["matrix"][clf.name] = metric_matrix

    return metrics_collection


def parse_multi_output(metrics_collection):

    result = {}
    for i, key in enumerate(list(metrics_collection.keys())):  # run level
        for metric in metrics_collection[key].keys():
            if i == 0:
                result[metric] = []
            result[metric].append(metrics_collection[key][metric])
    return result


def calculate_metrics_classifier(clf, X_test, y_test, y_pred):
    """
    The calculate_metrics_classifier function calculates the accuracy, average precision, precision and recall of a classifier.
    It takes as input two lists: y_test and y_pred. The function returns a dictionary with four keys: 'accuracy', 'average_precision',
    'precision' and 'recall'. Each key contains the corresponding value for each metric.

    :param y_test: Used to define the true labels of the test data.
    :param y_pred: Used to Calculate the metrics.
    :return: A dictionary with the following keys: accuracy, avgerage_preciscion, precision, and roc_auc.

    :doc-author: Julian M. Kleber
    """
    cf_matrix = confusion_matrix(y_test, y_pred)
    true_neg, false_pos, false_neg, true_pos = cf_matrix.ravel()
    acc = calculate_metric(y_test, y_pred, accuracy_score)
    prec = precision(true_pos, false_pos)
    rec = recall(true_pos, false_neg)
    spec = specificity(true_neg, false_pos)
    f_sc = f_score(prec, rec)
    fpr, tpr, thresholds = calculate_roc_curve(clf, X_test, y_test)
    auc_score = auc(1 - np.array(fpr), 1 - np.array(tpr))
    logging.info("Calculated metrics for %s" % clf.name)
    result_scalar = make_result(
        accuracy=acc,
        precision=prec,
        recall=rec,
        specificity=spec,
        f_score=f_sc,
    )
    result_array = make_result(
        fpr=fpr,
        tpr=tpr,
        roc_auc=auc_score,  # not an array but necessary for roc-auc plot
        thresholds=thresholds,
    )
    result_matrix = make_result(confusion_matrix=cf_matrix)
    return result_scalar, result_array, result_matrix


def calculate_metric(y_test, y_pred, metric):
    """
    The calculate_metric function takes as input the true labels and predicted labels
    and returns the specified metric. The available metrics are: accuracy, precision, recall, fscore.

    :param y_test: Used to pass the actual values of y.
    :param y_pred: Used to calculate the metric.
    :param metric: Used to specify the function that should be used to calculate the metric.
    :return: The result of the metric function.

    :doc-author: Julian M. Kleber
    """

    try:
        res = metric(y_test, y_pred)
    except:
        res = None
    return res


def make_result(**kwargs):

    """
    The make_result function takes in the accuracy, average precision, precision and roc_auc score of a model.
    It then returns a dictionary with these four values.

    :param acc: used to store the accuracy of the model.
    :param avg_prec: used to Store the average precision score of the model.
    :param prec: used to store the precision for each class.
    :param rac: sed to calculate the area under the curve.
    :return: A dictionary with four keys: accuracy, avgerage_preciscion, precision, and roc_auc.

    :doc-author: Trelent
    """
    return kwargs


def calculate_roc_curve(classifier, X_test, y_test):

    """
    The calculate_roc_curve function calculates the false positive rate and true positive rate for a given classifier.
    It returns these values as numpy arrays.

    :param classifier: Used to pass the classifier object.
    :param X_test: Used to test the model.
    :param y_test: Used to calculate the true positive rate and false positive rate.
    :return: The false positive rate, true positive rate and thresholds.

    :doc-author: Julian M. Kleber
    """

    probs = classifier.predict_proba(X_test)
    # Reading probabilities of the first class can extend it to all classes.
    probs = probs[:, 0]
    mean_fpr = np.linspace(0, 1, 100)
    fpr, tpr, thresholds = roc_curve(y_test, probs)
    interp_tpr = np.interp(mean_fpr, fpr, tpr)
    interp_fpr = np.interp(mean_fpr, tpr, fpr)
    interp_tpr = [0.0]+ interp_tpr.tolist()
    interp_fpr = [0.0]+ interp_fpr.tolist()
    return interp_fpr, interp_tpr, thresholds.tolist()


def precision(true_pos, false_pos):

    """
    The precision function takes two parameters: true_pos and false_pos.
    It returns the precision of the classifier, which is defined as
    the ratio of true positives (true_pos) to the sum of true positives and false positives (true_pos + false_pos).

    :param true_pos: Used to Calculate the precision.
    :param false_pos: Used to Calculate the precision.
    :return: A value between 0 and 1.

    :doc-author: Trelent
    """

    return true_pos / (true_pos + false_pos)


def recall(true_pos, false_neg):

    """
    The recall function takes two inputs: true positives and false negatives.
    It returns the recall of those values as a percentage.

    :param true_pos: Used to Calculate the recall.
    :param false_neg: Used to Calculate the recall.
    :return: The ratio of true positives to the sum of true positives and false negatives.

    :doc-author: Julian M. Kleber
    """

    return true_pos / (true_pos + false_neg)


def specificity(true_neg, false_pos):

    """
    The specificity function takes in the number of true negatives and false positives,
    and returns the specificity score for a given classifier

    :param true_neg: Used to Calculate the specificity.
    :param false_pos: Used to Calculate the specificity.
    :return: The proportion of negatives that are correctly identified as such (e.

    :doc-author: Trelent
    """

    return true_neg / (true_neg + false_pos)


def f_score(precision, recall):
    """
    The f_score function computes the harmonic mean of precision and recall.

    Args:
        precision (float): The number of true positives divided by all positive predictions.

        recall (float): The number of true positives divided by the number of positive values in the dataset.

    Returns:
        float: The f_score or accuracy for this prediction and target value, given as a floating point value between 0 and 1.

    :param precision: Used to Control the number of false positives.
    :param recall: Used to Avoid false positives.
    :return: The f_score of the precision and recall.

    :doc-author: Julian M. Kleber
    """

    return 2 * (precision * recall) / (precision + recall)


def init_metrics_collection() -> dict:
    """

    The init_metrics_collection function initializes the metrics_collection dictionary.
    The metrics_collection dictionary is a collection of dictionaries, each of which contains
    a set of keys and values. The first level key is "scalar", "array", or "matrix". The value for this key
    is another dictionary that contains the metric names as keys and their corresponding values as values.

    :return: A dictionary with three keys: 'scalar', 'array', 'matrix'.

    :doc-author: Julian M. Kleber
    """

    metrics_collection = {}
    metrics_collection["scalar"] = {}
    metrics_collection["array"] = {}
    metrics_collection["matrix"] = {}

    return metrics_collection



from rdkit.Chem import Draw


def plot_compound_list(mols, file_nam):
    """
    The plot_compound_list function takes a list of RDKit molecules and plots them in a grid.
    The function takes two arguments:
        mols - A list of RDKit molecules to plot.
        file_name - The name of the image file to save the plot as, including the extension (.png, .svg, etc.).

    :param mols: Used to specify the list of molecules to be drawn.
    :param file_nam: Used to name the image file.
    :return: A plot of the molecules in a list.

    :doc-author: Julian M. Kleber
    """

    drawOptions = Draw.rdMolDraw2D.MolDrawOptions()
    drawOptions.prepareMolsBeforeDrawing = False
    img = Draw.MolsToGridImage(
        mols,
        molsPerRow=5,
        subImgSize=(200, 200),
        legends=[
            x.GetProp("ReadyBiodegradability") + "    " + x.GetProp("Dataset")
            for x in mols
        ],
    )
    img.save(file_name)
    raise NotImplementedError("Not yet tested")



import numpy as np


def get_data_impurities(fitted_clf, X_data, y_true, n=6) -> list:
    """
    The get_data_impurities function takes a classifier, the data and true labels. It returns two lists of column names:
    the first list is the false positive columns, and the second list is the false negative columns. The function also takes an
    optional argument n which defaults to 6. This means that if there are more than 6 impurities in either category (false positives or negatives) then only
    the first six will be returned.

    :param clf: Used to store the classifier that is used to make predictions.
    :param X_data: Used to get the data that is used to train and test the model.
    :param y_true: Used to Get the true labels of the data.
    :param n=6: Used to Specify the number of false positive and negative values to be returned.
    :return: A list of two lists.

    :doc-author: Trelent
    """

    false_pos_col = []
    false_neg_col = []
    for i in range(n):
        false_neg, false_pos = get_false_predictions(
            fitted_clf=fitted_clf, X_data=X_data, y_true=y_true
        )
        false_pos_col.append(false_pos)
        false_neg_col.append(false_neg)
    false_pos = np.unique(false_pos_col).tolist()
    false_neg = np.unique(false_neg_col).tolist()
    return [false_pos, false_neg]


def get_false_predictions(fitted_clf, X_data, y_true) -> list:
    """
    The get_false_predictions function returns a tuple of two lists. The first list contains the indices of false positive predictions, and the second list contains the indices of false negative predictions.

    Parameters: clf (sklearn classifier object) - A trained sklearn classifier object
                X_data (numpy array) - A numpy array containing feature data for each example in dataset
                y_data (numpy array) - A numpy array containing labels for each example in dataset

                Returns: tuple(list): Two lists, one with all indices where a false positive occured, and one with all indicies where a false negative occured

    :param clf: Used to store the classifier that is being used.
    :param X_data: Used to store the data that is used to make predictions with the classifier.
    :param y_data: Used to get the true values of the data.
    :return: A tuple containing two lists: false_pos and false_neg.

    :doc-author: Julian M. Kleber
    """

    y_pred = fitted_clf.predict(X_data)
    mask_1 = y_true == 0
    mask_2 = y_pred == 1
    false_pos = np.where(mask_1 * mask_2)[0]
    mask_1 = y_true == 1
    mask_2 = y_pred == 0
    false_neg = np.where(mask_1 * mask_2)[0]

    return false_pos.tolist(), false_neg.tolist()



from sklearn.neural_network import MLPClassifier


class MLPClassifierAlgorithm(MLPClassifier):
    name = "mlp"
    hyperparameters = [
        {
            "alpha": [0.0001, 0.00005, 0.00001, 0.0000001],
            "learning_rate": ["constant", "inv_scaling", "adaptive"],
            "early_stopping": [True, False],
        }
    ]


hidden_layer_sizes = (100, 20, 20, 100)
max_iter = 10000
NeuralNetworkClassifierRELU = MLPClassifierAlgorithm(
    hidden_layer_sizes=hidden_layer_sizes,
    activation="relu",
    solver="adam",
    max_iter=max_iter,
)
NeuralNetworkClassifierRELU.name = "mlpra"

NeuralNetworkClassifierTanh = MLPClassifierAlgorithm(
    hidden_layer_sizes=hidden_layer_sizes,
    activation="tanh",
    solver="adam",
    max_iter=max_iter,
)
NeuralNetworkClassifierTanh.name = "mlpta"



# want to do with ROC check: https://scikit-learn.org/stable/auto_examples/model_selection/plot_roc_crossval.html#sphx-glr-auto-examples-model-selection-plot-roc-crossval-py
# making own scoring: https://scikit-learn.org/stable/modules/cross_validation.html
# Want to do Stratified k-fold
from sklearn.model_selection import GridSearchCV
from sklearn.model_selection import StratifiedKFold
import numpy as np
from sklearn.metrics import accuracy_score

# algorithm

# cross validation parameters
import logging


logging.basicConfig(
    format="%(levelname)s:%(asctime)s %(message)s",
    datefmt="%m/%d/%Y %I:%M:%S %p",
    level=logging.INFO,
    filename="chembee_actions.log",
)


def screen_cross_validation_grid_search(
    scores, clf_list, X_train, X_test, y_train, y_test, refit
) -> tuple:

    fitted_clf = {}
    result_clf = []
    for alg in clf_list:
        logging.info("Performing Grid search CV for" + str(alg.name))
        clf, report, best_parameters, best_index = cross_validation_grid_search(
            scores, alg, X_train, X_test, y_train, y_test, refit=refit
        )

        result = {
            "report": report,
            "best_parameters": best_parameters,
            "best_index": best_index,
        }
        fitted_clf[alg.name] = result
        result_clf.append(clf)
    return fitted_clf, result_clf


def cross_validation_grid_search(
    scores: list, clf, X_train, X_test, y_train, y_test, refit
) -> tuple:
    """
    The cross_validation_grid_search function takes in a list of scores, the classifier,
    the training data and labels, the test data and labels. It then performs a grid search
    on all hyperparameters for each score in scores using cross-validation on the training set.
    It returns a dictionary with keys as each score name and values as another dictionary with
    keys 'best_params' (a list of best parameters) 'best_score' (the best average cross-validation
    score), 'train_accuracy' (training accuracy on entire training set), 'test_accuracy'(testing accuracy
    on test set). The function also prints out all results.

    :param scores:list: Used to Specify the metrics that we want to use for evaluating our model.
    :param clf: Used to Pass the classifier that will be used in cross validation.
    :param X_train: Used to Train the algorithm.
    :param X_test: Used to Test the model on unseen data.
    :param y_train: Used to Fit the model.
    :param y_test: Used to Calculate the accuracy of the model.
    :param refit: Used to Choose the best model from the grid search.
    :return: The best parameters and the best estimator.

    :doc-author: Trelent
    """

    clf = GridSearchCV(clf, clf.hyperparameters, scoring=scores, n_jobs=-1, refit=refit)
    clf.fit(X_train, y_train)

    return clf, clf.cv_results_, clf.best_params_, clf.best_index_


def stratified_n_fold(clf, X_data, y_data, n=5, cut_off_filter=None) -> dict:
    """
    The stratified_n_fold function takes a classifier, training data and labels, and returns the average accuracy of the classifier on each fold.
    The function also takes an optional cut_off_filter parameter that allows you to filter out features with low variance.

    :param clf: Used to pass in the classifier that will be used.
    :param X_data: Used to pass the data to be used for training and testing.
    :param y_data: Used to specify the labels of the data.
    :param n=5: Used to specify the number of folds.
    :param cut_off_filter=None: Used to filter the data.
    :return: A dictionary with two lists: accuracies_train and accuracies_test.

    :doc-author: Julian M. Kleber
    """
    # Run classifier with cross-validation and plot ROC curves
    cv = StratifiedKFold(n_splits=n)
    accuracies_test = []
    accuracies_train = []
    for i, (train, test) in enumerate(cv.split(X_data, y_data.astype(np.int32))):

        clf.fit(X_data[train], y_data.astype(np.int32)[train])
        y_pred = clf.predict(X_data[test])
        score = accuracy_score(y_pred, y_data.astype(np.int32)[test])
        accuracies_test.append(float(score))
        y_pred = clf.predict(X_data[train])
        score = accuracy_score(y_pred, y_data.astype(np.int32)[train])
        accuracies_train.append(float(score))

    result = {"accuracies_train": accuracies_train, "accuracies_test": accuracies_test}

    return result


def stratified_n_fold_filter(clf, X_data, y_data, n=5, cut_off_filter=None) -> dict:
    """
    The stratified_n_fold_filter function takes in a classifier, training data and labels, and returns the average accuracy of the classifier on each fold. Additionally it also returns a list of indices that were filtered out due to having an accuracy below some cut_off_filter value.
    Note: To be SOLID and avoid weird code as well as performance issues the function, stratified_n_fold is basically implemented twice. One with and one time without a filter.
    :param clf: Used to pass the classifier that should be used.
    :param X_data: Used to pass the data to be used for training and testing.
    :param y_data: Used to determine the number of classes.
    :param n=5: Used to specify the number of folds in the stratifiedkfold function.
    :param cut_off_filter=None: Used to filter out the indices of the test set that have a lower accuracy than cut_off_filter.
    :return: A dictionary with the following keys:.

    :doc-author: Julian M. Kleber
    """

    cv = StratifiedKFold(n_splits=n)
    accuracies_test = []
    accuracies_train = []
    filtered_indices = []
    for i, (train, test) in enumerate(cv.split(X_data, y_data.astype(np.int32))):
        clf.fit(X_data[train], y_data.astype(np.int32)[train])
        y_pred = clf.predict(X_data[test])
        score = accuracy_score(y_pred, y_data.astype(np.int32)[test])
        accuracies_test.append(float(score))
        y_pred = clf.predict(X_data[train])
        score = accuracy_score(y_pred, y_data.astype(np.int32)[train])
        accuracies_train.append(float(score))
        if cut_off_filter != None:
            if score < cut_off_filter:
                filtered_indices.append(test)
    if len(filtered_indices) < 1:
        filtered_indices = None
    result = {
        "accuracies_train": accuracies_train,
        "accuracies_test": accuracies_test,
        "filtered_indices": filtered_indices,
    }

    return result



import pandas as pd
from mordred import Calculator, descriptors
import numpy as np

from rdkit import Chem
from rdkit.Chem import (
    Descriptors,
    Descriptors3D,
    rdMolDescriptors,
    PandasTools,
    RDKFingerprint,
)
from rdkit.Chem.AllChem import GetMorganFingerprintAsBitVect

# Utilities
import os
from tqdm import tqdm


def calculate_mordred_descriptors(mols: list) -> pd.DataFrame:
    """
    The calculate_mordred_descriptors function calculates Mordred descriptors for a set of molecules.
    The function takes as input a list of RDKit molecule objects and returns the calculated descriptors in
    a pandas dataframe.

    :param mols: Used to specify the molecules for which descriptors are to be calculated.
    :param : Used to specify the calculation of 3d descriptors.
    :return: A pandas dataframe containing the calculated descriptors.

    :doc-author: Julian M. Kleber
    """

    calc = Calculator(descriptors, ignore_3D=False)
    data = calc.pandas(mols)
    return data


def calculate_lipinski_desc(data_set: pd.DataFrame, mols: pd.Series) -> pd.DataFrame:
    """
    The calculate_lipinski_desc function calculates the molecular descriptors for each molecule in a data set.
    The function takes two arguments:
        1) data_set - A pandas DataFrame containing the SMILES strings of each molecule in a data set.
        2) mols - A pandas Series containing the RDKit Mol object of each molecule in a data set.

    The function returns one value:
        1) The original DataFrame with additional columns for each descriptor, which contain floats.

    :param data_set:pd.DataFrame: Used to Store the calculated descriptors.
    :param mols:pd.Series: Used to Pass a series of molecules to the function.
    :return: A dataframe with the descriptors for each molecule in the mols series.

    :doc-author: Trelent
    """

    for i, mol in tqdm(enumerate(mols)):
        Chem.SanitizeMol(mol)
        data_set.loc[i, "Molecule"] = mol
        data_set.loc[i, "MolWt"] = Descriptors.MolWt(mol)
        data_set.loc[i, "LogP"] = Descriptors.MolLogP(mol)
        data_set.loc[i, "NumHAcceptors"] = Descriptors.NumHAcceptors(mol)
        data_set.loc[i, "NumHDonors"] = Descriptors.NumHDonors(mol)
        data_set.loc[i, "NumHeteroatoms"] = Descriptors.NumHeteroatoms(mol)
        data_set.loc[i, "NumRotatableBonds"] = Descriptors.NumRotatableBonds(mol)
        data_set.loc[i, "NumHeavyAtoms"] = Descriptors.HeavyAtomCount(mol)
        data_set.loc[
            i, "NumAliphaticCarbocycles"
        ] = Descriptors.NumAliphaticCarbocycles(mol)
        data_set.loc[
            i, "NumAliphaticHeterocycles"
        ] = Descriptors.NumAliphaticHeterocycles(mol)
        data_set.loc[i, "NumAliphaticRings"] = Descriptors.NumAliphaticRings(mol)
        data_set.loc[i, "NumAromaticCarbocycles"] = Descriptors.NumAromaticCarbocycles(
            mol
        )
        data_set.loc[
            i, "NumAromaticHeterocycles"
        ] = Descriptors.NumAromaticHeterocycles(mol)
        data_set.loc[i, "NumAromaticRings"] = Descriptors.NumAromaticRings(mol)
        data_set.loc[i, "RingCount"] = Descriptors.RingCount(mol)
        data_set.loc[i, "FractionCSP3"] = Descriptors.FractionCSP3(mol)

        data_set.loc[i, "TPSA"] = Descriptors.TPSA(mol)
        data_set.loc[i, "NPR1"] = rdMolDescriptors.CalcNPR1(mol)
        data_set.loc[i, "NPR2"] = rdMolDescriptors.CalcNPR2(mol)
        data_set.loc[i, "InertialShapeFactor"] = Descriptors3D.InertialShapeFactor(mol)
        data_set.loc[i, "RadiusOfGyration"] = Descriptors3D.RadiusOfGyration(mol)
    return data_set


def convert_mols_to_rdk_fp(mols, return_bit=False):
    """
    The convert_mols_to_rdkfingerprint function converts a list of RDKit molecules into a list of RDKit fingerprints.

    :param mols: Used to pass the molecules that are to be converted into fingerprints.
    :return: A list of rdkit fingerprints.

    :doc-author: Julian M. Kleber
    """
    if return_bit:
        fps = [np.array(RDKFingerprint(x)).tolist() for x in mols]
    else:
        fps = [RDKFingerprint(x) for x in mols]
    return fps


def convert_mols_to_morgan_fp(mols, radius=3, n_bits=2048, return_bit=False):
    """
    The convert_mols_to_morgan_fp function converts a list of molecules to their Morgan fingerprints.

    :param mols: Used to specify the molecules that are to be converted into Morgan fingerprints.
    :param radius=3: Used to determine the number of bits that will be used to represent each molecule.
    :param n_bits=2048: Used to determine the number of bits in the Morgan fingerprint.
    :return: A list of fingerprints.

    :doc-author: Julian M. Kleber
    """
    if return_bit:
        fps = [
            GetMorganFingerprintAsBitVect(
                np.array(x).tolist(),
                useChirality=True,
                useFeatures=True,
                radius=radius,
                nBits=n_bits,
            )
            for x in mols
        ]
    else:
        fps = [
            GetMorganFingerprintAsBitVect(
                x, useChirality=True, useFeatures=True, radius=radius, nBits=n_bits
            )
            for x in mols
        ]
    return fps


def convert_mol_to_inchi(mols):
    """
    The convert_mol_to_inchi function converts a molecule into an InChI string.

    :param mol: Used to pass the molecule to the function.
    :return: A tuple of the inchi and inchikey.

    :doc-author: Julian M. Kleber
    """

    from rdkit import Chem

    inchis = []
    inchikeys = []
    for mol in mols:
        inchis.append(str(Chem.MolToInchi(mol)))
        inchikeys.append(str(Chem.MolToInchiKey(mol)))

    return inchis, inchikeys


def get_mols_from_supplier(indices, supplier):
    """
    The get_mols_from_supplier function takes a list of indices and a supplier object,
    and returns the molecules corresponding to those indices.

    :param indices: Used to specify which molecules to retrieve from the supplier.
    :param supplier: Used to specify the supplier of the molecules.
    :return: A list of molecules.

    :doc-author: Julian M. Kleber
    """

    mols = []
    for indx in indices:
        mol = supplier[indx]
        mols.append(mol)
    return mols


def load_data(file_path: str):
    """
    The load_data function loads the data from a file and returns it as a Pandas DataFrame.

    :param file_path:str: Used to specify the location of the.
    :return: A dataframe with the columns specified by the sdf data entries

    :doc-author: Julian M. Kleber
    """

    sdfFile = os.path.join(file_path)
    mols = Chem.SDMolSupplier(sdfFile)
    frame = PandasTools.LoadSDF(
        sdfFile,
        smilesName="SMILES",
        molColName="Molecule",
        includeFingerprints=True,
        removeHs=False,
        strictParsing=True,
    )
    frame = calculate_lipinski_desc(data_set=frame, mols=mols)
    return frame


if __name__ == "__main__":
    from utils import save_csv

    data = load_data("tests/data/Biodeg.sdf")
    save_csv(data, "converted_data.csv")



import matplotlib.pyplot as plt
from file_utils import insert_string_in_file_name, prepare_file_name_saving
import numpy as np


def plot_feature_importances(
    result_json: dict, file_name: str, prefix: str, show_x_label=True
):
    """
    The plot_feature_importances function plots the feature importances of a random forest model.
    It takes three arguments:
        - forest_importances: The feature importances from the random forest model, as returned by sklearn's .feature_importance() method.
        - std: The standard deviation of the feature importances, as returned by sklearn's .std() method on a RandomForestRegressor or RandomForestClassifier object.  This is used to plot error bars for each importance value.  If no standard deviation is available (i.e., if you are using a DecisionTreeRegressor or DecisionTreeClassifier), set this to None (the default).
        - file_name: A string containing the name of your desired output file (e.g., "FeatureImportancePlot" will result in "FeatureImportancePlotRFModelNameHere").
        - show_y_label: Boolean to indicate if there should be a label for the y axis.
    :param forest_importances: Used to pass the feature importances to the function.
    :param std: Used to plot the standard deviation of the feature importances.
    :param file_name:std: Used to pass the file name of the plot to be saved.
    :param prefix:std: Used to specify the prefix for the file name.
    :return: The feature importances of the forest.

    :doc-author: Julian M. Kleber
    """
    importances = result_json["importances"]
    feature_indices = result_json["feature_indices"]

    std = result_json["std"]

    file_name = insert_string_in_file_name(
        file_name, insertion="FeatureImportancesRF", ending=".png"
    )
    file_name = prepare_file_name_saving(
        prefix=prefix, file_name=file_name, ending=".png"
    )
    fig, ax = plt.subplots()
    ax.bar(x=feature_indices, height=importances, yerr=std)
    ax.set_xlabel("Feature No.")
    ax.set_ylabel("Mean decrease in impurity")
    if show_x_label == False:
        ax.set_xticklabels(
            np.arange(1, len(feature_indices) + 1, int(len(feature_indices) / 10))
        )
        ax.set_xticks(
            np.arange(1, len(feature_indices) + 1, int(len(feature_indices) / 10))
        )
    fig.tight_layout()
    fig.savefig(file_name, dpi=300)
    plt.cla()
    plt.clf()
    plt.close()



from file_utils import prepare_file_name_saving


class DataSet:

    # raw data (don't touch)
    name = None
    data = None

    def load_data_set():
        raise NotImplementedError

    def make_train_test_split(self, data_set, split_ratio):
        raise NotImplementedError

    def get_split(self):

        return self.X_train, self.X_test, self.y_train, self.y_test

    def save_data_csv(self, file_name, prefix=None):

        file_name = prepare_file_name_saving(prefix=prefix, file_name=file_name)
        self.data.to_csv(file_name)



from sklearn.linear_model import LogisticRegression


class LogisticRegressionClassifierAlgorithm(LogisticRegression):

    name = "logr"


LogisticRegressionClassifier = LogisticRegressionClassifierAlgorithm(
    multi_class="multinomial"
)



import numpy as np


def get_top_n_similar_compounds(similarities: list, n: int = 10, labels=None):
    """
    The get_top_n_similar_compounds function takes in a list of similarities and returns the top n similar compounds.
    The function takes in a list of similarities, sorts it, and then returns the top n most similar compounds.

    :param similarities:list: Used to pass in the list of similarities for each compound.
    :param n:int=10: Used to Specify the number of similar compounds to return.
    :return: The top n similar compounds for a given compound, for each similarity list

    :doc-author: Julian M. Kleber
    """
    if labels != None:
        assert len(labels) == len(similarities)
    similar_compounds_screen = {}
    for i, sims in enumerate(similarities):
        similar_compounds = np.argpartition(similarities, -n)[-n:].tolist()
        if labels != None:
            similar_compounds_screen[labels[i]] = labels
    return similar_compounds_screen


def get_similar_compounds(
    compounds_of_interest: np.ndarray,
    compounds: np.ndarray,
    distance="tanimoto",
    return_sim_matrix=True,
) -> dict:
    """
    The get_similar_compounds function takes a list of compounds and returns the n most similar compounds.
    The similarity is calculated by calculating a user defined distance metric between two compound vectors.
    The function works on arbitrary data and features. It could also take in vectorized fingerprints. The function
    is written as such that it could return the most similar compounds that are the same as the search compounds. You
    would have to exclude them manually. Still, the function is experimental and if you feel handling the exlusion of
    of identical compounds inside of this function is neccessary, please file an issue, or feature request.

    :param compounds_of_interest: Used to specify the compounds that we want to find similar compounds for.
    :param compounds: Used to store the similarity scores for each compound.
    :param n=10: Used to specify the number of similar compounds to return. Specify n='all' if you want to return all coumpounds
    :return: A dictionary containing the jsonified result, ready to use in web tech like MongoDB, Flask, React, etc., and another dictionary cointaining the complete similarity matrix

    :doc-author: Julian M. Kleber
    """
    from pyADA import ApplicabilityDomain

    AD = ApplicabilityDomain(verbose=True)
    sims = AD.analyze_similarity(
        base_train=compounds,
        base_test=compounds_of_interest,
        similarity_metric=distance,
    )
    if return_sim_matrix:
        return sims.to_dict(), AD.similarities_table_.to_dict()
    else:
        return sims.to_dict()


def screen_fingerprints_against_data(to_screen, base) -> dict:
    """
    The screen_fingerprints_against_data function takes two arguments: a list of fingerprints to screen, and a list of
    fingerprints against which the first argument will be screened. The function returns a dictionary with keys that are
    the fingerprint names in the first argument, and values that are lists containing the similarity scores for each molecule
    in the second argument. For example:

        >>>screen_fingerprints_against_data([FP11, FP12], [FP21, FP22])

        {1: {similarityScores: [0.24, 0.6], fingerPrint: FP11}, 2: {similarityScores: [0.1, 0.42], fingerPrint: FP12}}

    :param to_screen: Used to specify the fingerprints to be screened against the base.
    :param base: Used to define the set of molecules to compare against.
    :return: A dictionary with the similarity scores of the fingerprints in to_screen against all of the fingerprints in base.

    :doc-author: Julian M. Kleber
    """
    from rdkit.DataStructs import FingerprintSimilarity

    result = {}
    for count, probe in enumerate(to_screen):
        result[count] = {}
        sim = []
        for mol in base:
            sim.append(FingerprintSimilarity(probe, mol))
        result[count]["simalarityScores"] = sim
        result[count]["fingerPrint"] = probe
    return result


def get_similar_compounds_structure(
    compounds_of_interest, compounds, n=10, distance="tanimoto"
) -> dict:
    """
    The get_similar_compounds_structure function takes a list of compounds and returns the most similar compounds based on structure.
    The function takes in four arguments:

        1) A list of compound names (compounds_of_interest),
        2) A dataframe containing all the other compounds (compounds), and
        3) An integer n that specifies how many similar compounds to return. The default is 10.
        4) A distnace

    The function returns a dictionary with three keys: "InChi", "number" and "similarity". Number corresponds to the index number for each compound, while similarity contains their similarity score.

    :param compounds_of_interest: Used to Specify the compounds for which similar compounds are to be found.
    :param compounds: Used to Get the structure of all compounds in the database.
    :param n=10: Used to Specify how many similar compounds we want to find.
    :param distance="Tanimoto": Used to Specify the similarity metric to be used.
    :return: A dictionary with the following keys:.

    :doc-author: Julian M. Kleber
    """
    raise NotImplementedError(
        "Either implement with RDKit or just convert a fingerprint to a dataframe"
    )
    for i in range(len(compounds_of_interest)):
        compound = compounds_of_interest[i]
        similarity = calculate_similarity(compound, compounds, n=10)
        compounds["number"] = i
        compounds["similarity"] = similarity
        compounds["InChI"] = inchi
    return compounds



import os
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))


from sklearn.ensemble import RandomForestClassifier


from chembee_config.benchmark.BenchmarkAlgorithm import BenchmarkAlgorithm


class RandomForestClassifier(BenchmarkAlgorithm):

    name = "random-forest"

    algorithms = (
        RandomForestClassifier(
            n_estimators=100,
            criterion="gini",
            max_depth=None,
            min_samples_split=2,
            min_samples_leaf=1,
            min_weight_fraction_leaf=0.0,
            max_features="sqrt",
            max_leaf_nodes=None,
            min_impurity_decrease=0.0,
            bootstrap=True,
            oob_score=False,
            n_jobs=None,
            random_state=None,
            verbose=0,
            warm_start=False,
            class_weight=None,
            ccp_alpha=0.0,
            max_samples=None,
        ),
        RandomForestClassifier(
            n_estimators=100,
            criterion="entropy",
            max_depth=None,
            min_samples_split=2,
            min_samples_leaf=1,
            min_weight_fraction_leaf=0.0,
            max_features="sqrt",
            max_leaf_nodes=None,
            min_impurity_decrease=0.0,
            bootstrap=True,
            oob_score=False,
            n_jobs=None,
            random_state=None,
            verbose=0,
            warm_start=False,
            class_weight=None,
            ccp_alpha=0.0,
            max_samples=None,
        ),
        RandomForestClassifier(
            n_estimators=100,
            criterion="log_loss",
            max_depth=None,
            min_samples_split=2,
            min_samples_leaf=1,
            min_weight_fraction_leaf=0.0,
            max_features="sqrt",
            max_leaf_nodes=None,
            min_impurity_decrease=0.0,
            bootstrap=True,
            oob_score=False,
            n_jobs=None,
            random_state=None,
            verbose=0,
            warm_start=False,
            class_weight=None,
            ccp_alpha=0.0,
            max_samples=None,
        ),
    )

    titles = (
        "Gini coefficient loss",
        "Shannon entropy loss",
        "Shannon log-loss",
    )



from chembee_datasets.DataSet import DataSet
import pandas as pd
from sklearn.model_selection import train_test_split
from rdkit import Chem
from rdkit.Chem import (
    PandasTools,
)

from file_utils import prepare_file_name_saving


class BioDegDataSet(DataSet):

    name = "biodeg"

    def __init__(self, data_set_path, target, split_ratio=0.7):

        self.data, self.mols = self.load_data_set(data_set_path)
        self.split_ratio = split_ratio
        (
            self.X_train,
            self.X_test,
            self.y_train,
            self.y_test,
        ) = self.make_train_test_split(self.data, self.split_ratio, y_col=target)

        self.feature_names = self.get_feature_names(self.data, target=target)

    def clean_data(self, data):
        """
        The clean_data function takes in a dataframe and cleans it by removing the SMILES, CASRN, ID and Dataset columns.
        It also converts all of the dtypes to float64 or int64. It returns a cleaned dataframe.

        :param self: Used to Reference the class itself.
        :param data: Used to Pass the data that is to be cleaned.
        :return: A dataframe with the columns "smiles", "dataset", "casrn" and "id" dropped, and all other columns converted to numeric type.

        :doc-author: Trelent
        """

        data = data.drop(columns=["SMILES", "Dataset", "CASRN", "ID"])
        data = data.convert_dtypes()
        bad_types = data.select_dtypes(
            exclude=["string", "int64", "float64"]
        ).columns.to_list()
        data = data.drop(columns=bad_types)
        return data

    def load_data_set(self, file_name: str):
        """
        The load_data function loads the data from a sdf file and returns it as a Pandas DataFrame.

        :param file_path:str: Used to Specify the location of the.
        :return: A dataframe with the following columns:.

        :doc-author: Trelent
        """
        mols = Chem.SDMolSupplier(file_name)
        frame = PandasTools.LoadSDF(
            file_name,
            smilesName="SMILES",
            molColName="Molecule",
            includeFingerprints=True,
            removeHs=False,
            strictParsing=True,
        )
        return frame, mols

    def load_data_set_from_csv(self, file_name: str) -> pd.DataFrame:
        """
        The load_data_set_from_csv function loads a csv file into a pandas dataframe.
        The function's purpose is to improve readability. Nothing special
        The function takes one argument, the name of the csv file to be loaded.
        The function returns a pandas dataframe containing all of the information in
        the specified csv file.

        :param self: Used to Access variables that belongs to the class.
        :param file_name:str: Used to Specify the name of the csv file that will be used to load data from.
        :return: A pandas data frame.

        :doc-author: Julian M. Kleber
        """

        return pd.from_csv(file_name)

    def make_train_test_split(self, data, split_ratio: float, y_col: str, shuffle=True):
        """
        The make_train_test_split function splits the data into a training set and test set.
        The split_ratio parameter determines how much of the data is used for training, and how much is used for testing.
        The y_col parameter specifies which column in the dataset contains labels (y). The shuffle parameter allows you to specify whether or not to shuffle your data before splitting it into train/test sets.

        :param self: Used to Reference the class itself.
        :param data: Used to Specify the dataframe that is used for splitting.
        :param split_ratio:float: Used to Determine the ratio of data that will be used for training.
        :param y_col:str: Used to Specify the column name of the dependent variable.
        :param shuffle=True: Used to Shuffle the data before splitting it into training and testing sets.
        :return: The X_train, X_test, y_train and y_test dataframes.

        :doc-author: Julian M. Kleber
        """

        X = data.drop([y_col], axis=1)
        y = data[y_col]
        train_samples = int(split_ratio * len(X))
        X_train, X_test, y_train, y_test = train_test_split(
            X,
            y,
            shuffle=shuffle,
            test_size=len(X) - train_samples,
        )
        return (
            X_train,
            X_test,
            y_train,
            y_test,
        )

    def save_data_csv(self, data: pd.DataFrame, file_name, prefix):
        """
        The save_data_csv function saves the data of the instance to a csv file.

        :param self: Used to Access the attributes and methods of the class in python.
        :param data:pd.DataFrame: Used to Save the data that is passed to it.
        :param file_name: Used to Specify the name of the file to be saved.
        :param prefix: Used to Add a prefix to the file name.
        :return: The name of the file where the data was saved.

        :doc-author: Julian M. Kleber
        """

        assert type(data) == type(pd.DataFrame()), "Data must be a pandas.DataFrame"
        file_name = prepare_file_name_saving(
            prefix=prefix, file_name=file_name, ending=".csv"
        )
        data.to_csv(file_name)

    def save_data_sdf(self, data, file_name, prefix, molColName="Molecule"):
        """
        The save_data_sdf function saves the data in a pandas DataFrame to an sdf file.

        :param self: Used to access the attributes and methods of the class in a method.
        :param data: Used to specify the dataframe that is to be saved.
        :param file_name: Used to specify the name of the file to which data is saved.
        :param prefix: Used to add a prefix to the file name.
        :param molColName="Molecule": Used to specify the name of the molecule column in the sdf file.
        :return: A dataframe with the same number of rows as the input data and one column named "romol" containing a mol object.

        :doc-author: Julian M. Kleber
        """

        file_name = prepare_file_name_saving(
            prefix=prefix, file_name=file_name, ending=".sdf"
        )
        PandasTools.WriteSDF(
            data, file_name, molColName="Molecule", properties=list(data.columns)
        )

    def get_feature_names(self, data, target):
        """
        The get_feature_names function takes in a dataframe and returns the feature names.
        It does this by iterating through the columns of the dataframe and appending them to a list.
        The function then returns that list.

        :param self: Used to reference the class itself.
        :param data: Used to drop the target column from the dataframe.
        :param target: Used to determine if the function is being used for a training or testing set.
        :return: The names of the features.

        :doc-author: Trelent
        """

        if not target:
            data = data.drop(columns=[target])
        return data.columns.to_list()



import matplotlib.pyplot as plt
import pandas as pd

from sklearn.inspection import DecisionBoundaryDisplay
from file_utils import prepare_file_name_saving

# set the standard parameters

plt.rcParams["font.size"] = "20"
plt.tight_layout()


def plot_comparison_4(
    models,
    titles,
    X,
    y,
    prefix=None,
    file_name="SVC_benchmark",
    feature_names=["Feature 1", "Feature 2"],
    response_method="predict",
):
    """
    The plot_comparision_4 function plots the decision boundaries for a set of models.
    It takes as input:
    - models: A list of scikit learn model objects.  Each model object must have a .predict method that takes in an array of features and returns predictions based on those features.  The plot_comparision_4 function uses these predictions to create the decision boundary for each model, and then plots them together in a single figure.
    - titles: A list with one element for each element in 'models'.  Each string is used to label the corresponding plot's legend, so make sure that they are unique!
    - X: An array containing feature values used to train/evaluate the models passed via 'models'.   These feature values should be scaled between 0 and 1 (if not already done).   The first two columns should correspond to x coordinates while the second two columns should correspond to y coordinates (i.e., this is an Nx4 matrix).  This argument corresponds directly with what you would pass into plt.scatter(...), so if you have already scaled your data, then just pass it here without re-scaling it!
    - y: An array containing labels corresponding with observations contained within X (i.e., this

    :param models: Used to Pass a list of models that should be compared.
    :param titles: Used to Give a name to each model in the plot.
    :param X: Used to Define the data that will be used for training and testing.
    :param y: Used to Define the target variable.
    :param prefix=None: Used to Define a string that will be added to the name of the file when it is saved.
    :param file_name="SVC_benchmark": Used to Save the plot as an image.
    :param feature_names=["Feature1": Used to Specify the labels for the x and y axis.
    :param "Feature2"]: Used to Specify which feature is used for the x-axis.
    :return: The subplot of the models.

    :doc-author: Trelent
    """

    fig, sub = plt.subplots(2, 2, figsize=(20, 13))
    plt.subplots_adjust(wspace=0.4, hspace=0.4)
    make_comparison_plot(
        fig=fig,
        sub=sub,
        models=models,
        titles=titles,
        X=X,
        y=y,
        file_name=file_name,
        feature_names=feature_names,
        prefix=prefix,
        response_method=response_method,
    )


def plot_comparison_3(
    models,
    titles,
    X,
    y,
    prefix=None,
    file_name="SVC_benchmark",
    feature_names=["Feature 1", "Feature 2"],
    response_method="predict",
):
    """
    The plot_comparison_3 function creates a 3x3 grid of plots, each comparing the performance of a different model.
    The models are passed in as an array and their titles as another array. The X and y data is also required to generate
    the plots. The function will automatically save the plot to file_name + .png in your current working directory unless
    otherwise specified using prefix = "". If you would like to save multiple comparisons, simply call this function once with
    different sets of parameters.

    :param models: Used to pass the models that are to be compared.
    :param titles: Used to set the title of each plot.
    :param X: Used to pass the data.
    :param y: Used to specify the target variable.
    :param prefix=None: Used to pass a string to the plot_comparison_3 function.
    :param file_name="SVC_benchmark": Used to name the file that will be saved in your directory.
    :param feature_names=["Feature1": Used to specify the names of the features in x.
    :param "Feature2"]: Used to Specify the name of the feature that will be plotted on the y-axis.
    :param response_method="predict": Used to specify the method used to generate the response.
    :param : Used to specify the type of response.
    :return: A dictionary with the trained models.

    :doc-author: Julian M. Kleber
    """

    fig, sub = plt.subplots(1, 3, figsize=(20, 13))
    plt.subplots_adjust(wspace=0.4, hspace=0.4)
    make_comparison_plot(
        fig=fig,
        sub=sub,
        models=models,
        titles=titles,
        X=X,
        y=y,
        file_name=file_name,
        feature_names=feature_names,
        prefix=prefix,
        response_method=response_method,
    )


def plot_comparison_2():
    raise NotImplementedError()


def plot_comparison_1(
    clf,
    X,
    y,
    prefix=None,
    file_name="SVC_benchmark",
    feature_names=["Feature 1", "Feature 2"],
    response_method="predict",
):
    """
    The plot_comparison_1 function creates a plot of the decision boundary for a given classifier.
    It takes as input:
    - clf, which is an instance of sklearn's SVC class with default parameters (except for C and gamma)
    - X, which is the feature matrix containing all training samples in rows and their features in columns.
    The number of columns must be equal to two because this function plots each sample as a point on the xy-plane.
    This function assumes that there are only two features in X (i.e., it can only plot 2D decision boundaries).
    If you want to use more than two features, then you need to reduce your data set so that it contains only 2D points!

        - y, which is the vector containing all labels corresponding to each training sample from X

        - prefix=None: A string indicating what should be printed before file_name when saving files; if None or empty string then no prefix will be added

        - file_name="SVC_benchmark": The name used when saving files; if None or empty string then no file will be saved but instead everything will simply get plotted on screen using plt.show()

        - feature_names=

    :param clf: Used to pass the classifier to be used.
    :param X: Used to specify the features that are used to train the model.
    :param y: Used to specify the class labels.
    :param prefix=None: Used to specify a prefix for the file name.
    :param file_name="SVC_benchmark": Used to save the plot as a png file.
    :param feature_names=["Feature1": Used to set the labels on the x and y axis.
    :param "Feature2"]: Used to set the name of the second feature.
    :param response_method="predict": Used to specify that the.
    :param : Used to set the title of the plot.
    :return: The decision boundary plot of the svc model.

    :doc-author: Julian M. Kleber
    """
    plt.rcParams["font.size"] = "35"
    X0, X1 = X[:, 0], X[:, 1]
    fig, ax = plt.subplots(1, 1, figsize=(20, 13))
    disp = DecisionBoundaryDisplay.from_estimator(
        clf,
        X,
        response_method=response_method,
        cmap=plt.cm.coolwarm,
        alpha=0.7,
        ax=ax,
        xlabel=feature_names[0],
        ylabel=feature_names[1],
    )
    plt.scatter(X0, X1, c=y, cmap=plt.cm.coolwarm, s=20, edgecolors="k")
    ax.set_xticklabels(())
    ax.set_yticklabels(())

    file_name = prepare_file_name_saving(
        prefix=prefix, file_name=file_name, ending=".png"
    )
    fig.savefig(file_name)
    plt.cla()
    plt.clf()
    plt.close()


def make_comparison_plot(
    fig,
    sub,
    models,
    titles,
    X,
    y,
    prefix=None,
    file_name="SVC_benchmark",
    feature_names=["Feature 1", "Feature 2"],
    response_method="predict",
):
    """
    The make_comparison_plot function creates a plot comparing the decision boundaries of several models.
    It takes as input:
        - fig, a matplotlib figure object to be saved to file (figsize=(12,8))
        - sub, an array of axes objects on which plotting will occur (nrows=2, ncols=3)
        - models: an array of sklearn classifiers for which decision boundaries will be plotted.
            Note that these must have been fitted already with fit()!

        The function then creates plots comparing the decision boundary and prediction values for each model in turn.

        It returns nothing.

    :param fig: Used to pass the figure object to which we want to plot.
    :param sub: Used to create a subplot grid.
    :param models: Used to pass the classifiers that will be used in the plot.
    :param titles: Used to set the title of each subplot.
    :param X: Used to pass the data to be plotted.
    :param y: Used to specify the response variable.
    :param prefix=None: Used to make sure that the file_name is saved in the current working directory.
    :param file_name="SVC_benchmark": Used to create a file_name for the plot.
    :param feature_names=["Feature1": Used to label the axes of the plot.
    :param "Feature2"]: Used to specify the name of the x and y axis.
    :return: A plot of the decision boundaries for a set of classifiers.

    :doc-author: Julian M. Kleber
    """

    X0, X1 = X[:, 0], X[:, 1]

    for clf, title, ax in zip(models, titles, sub.flatten()):
        disp = DecisionBoundaryDisplay.from_estimator(
            clf,
            X,
            response_method=response_method,
            cmap=plt.cm.coolwarm,
            alpha=0.7,
            ax=ax,
            xlabel=feature_names[0],
            ylabel=feature_names[1],
        )
        ax.scatter(X0, X1, c=y, cmap=plt.cm.coolwarm, s=20, edgecolors="k")
        ax.set_xticks(())
        ax.set_yticks(())
        ax.set_title(title)

    file_name = prepare_file_name_saving(
        prefix=prefix, file_name=file_name, ending=".png"
    )
    fig.savefig(file_name)
    plt.cla()
    plt.clf()
    plt.close()


def save_fig(ax, name: str, save_path=None, dpi=300):
    """Frequently used snippet to save a plot

    Args:
        ax ([type]): Seaborn axis
        name (str): Name of the plot
        save_path ([type], optional): Where to save the plot. Defaults to None:str.
        dpi (int, optional): Resolution of the plot. Defaults to 300.

    Returns:
        (None)
    """
    if save_path is None:
        save_path = name
    else:
        save_path.joinpath(name)
    fig = ax.get_figure()
    fig.savefig(save_path, dpi=dpi)
    plt.cla()
    plt.clf()
    plt.close()
    return None


plotting_map_comparison = {
    1: plot_comparison_1,
    2: plot_comparison_2,
    3: plot_comparison_3,
    4: plot_comparison_4,
}
if __name__ == "__main__":

    data = pd.read_csv("converted_data.csv")
    biodeg = data[data["ReadyBiodegradability"] == 1]
    n_biodeg = data[data["ReadyBiodegradability"] == 0]
    polar_plot(biodeg, name="BiodegPolar.png")
    polar_plot(n_biodeg, name="NBiodegPolar.png")



from chembee_datasets.BreastCancer import BreastCancerDataset
from sklearn import datasets


class IrisDataSet(BreastCancerDataset):

    name = "iris-dataset"

    def load_data_set(self):

        iris_data = datasets.load_iris()
        return iris_data



from sklearn.linear_model import LinearRegression


class LinearRegressionClass(LinearRegression):

    name = "linr"

    def fit(self, X, y):
        super().fit(X, y)
        df = self.decision_function(X)
        self.df_min_ = df.min()
        self.df_max_ = df.max()

    def predict_proba(self, X):
        """Min-max scale output of `decision_function` to [0,1]."""
        df = self.decision_function(X)
        calibrated_df = (df - self.df_min_) / (self.df_max_ - self.df_min_)
        proba_pos_class = np.clip(calibrated_df, 0, 1)
        proba_neg_class = 1 - proba_pos_class
        proba = np.c_[proba_neg_class, proba_pos_class]
        return proba

    def decision_function(self, X):

        return self.intercept_ + self.coef_ * X


LinearRegressionClassifier = LinearRegressionClass()



from pyADA import ApplicabilityDomain
import numpy as np


def get_applicability_domain(
    clf,
    X_train: np.ndarray,
    X_test: np.ndarray,
    y_train: np.ndarray,
    y_test: np.ndarray,
    threshold_step=(0, 0.65, 0.05),
    similarity_metric="tanimoto",
    metric_evaluation="auc",
) -> dict:
    """
    The get_applicability_domain function takes as input a classifier, training and test data,
    and returns the applicability domain of the classifier. The function also takes in optional arguments
    for threshold_step (default is 0.05), similarity metric (default is tanimoto), and metric evaluation
    (default is AUC). The function outputs a dictionary with keys 'threshold', 'similarity', and 'metric'
    with corresponding values for each key. It depends on the pyADAqsar package and further documentation can be found there
    (https://github.com/jeffrichardchemistry/pyADA)

    :param clf: Used to Specify the classifier that will be used to evaluate the applicability domain.
    :param X_train: Used to Train the model.
    :param X_test: Used to Define the test set to be used for applicability domain construction.
    :param y_train: Used to Fit the model on.
    :param y_test: Used to Evaluate the performance of the classifier.
    :param threshold_step=(0: Used to Define the step size of the threshold.
    :param 0.65: Used to Define the threshold for the applicability domain.
    :param 0.05): Used to Define the step size of the threshold.
    :param similarity_metric="tanimoto": Used to Define the similarity metric that will be used to compute the applicability domain.
    :param metric_evaluation="auc": Used to Specify the metric used to evaluate the applicability domain.
    :param : Used to Define the threshold step.
    :return: A dictionary with the following keys:.

    :doc-author: Julian M. Kleber
    """

    ad = ApplicabilityDomain(verbose=True)
    result = ad.fit(
        model=clf,
        base_test=X_test,
        base_train=X_train,
        y_true=y_test,
        threshold_reference="max",
        threshold_step=threshold_step,
        similarity_metric="tanimoto",
        alpha=1,
        beta=1,
        metric_avaliation="auc",
    )
    return result



import logging

LOGGER = logging.getLogger(__name__)


def clf_fit(clf, X_train, y_train, name):
    """
    The clf_fit function fits the classifier to the training data. The function serves the only purpose of readability and maintainability-

    :param X_train: Used to Pass the x_train dataset to the clf.
    :param y_train: Used to Fit the model to the training data.
    :param name: Used to Name the classifier.
    :return: The fitted model.

    :doc-author: Julian M. Kleber
    """

    LOGGER.info("Fitting %s" % (name))
    clf.fit(X_train, y_train)
    y_pred = clf.predict(X_train)
    return clf, y_pred
